
#ifndef PDF4QTLIBWIDGETSSHARED_EXPORT_H
#define PDF4QTLIBWIDGETSSHARED_EXPORT_H

#ifdef PDF4QTLIBWIDGETS_STATIC_DEFINE
#  define PDF4QTLIBWIDGETSSHARED_EXPORT
#  define PDF4QTLIBWIDGETS_NO_EXPORT
#else
#  ifndef PDF4QTLIBWIDGETSSHARED_EXPORT
#    ifdef Pdf4QtLibWidgets_EXPORTS
        /* We are building this library */
#      define PDF4QTLIBWIDGETSSHARED_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define PDF4QTLIBWIDGETSSHARED_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef PDF4QTLIBWIDGETS_NO_EXPORT
#    define PDF4QTLIBWIDGETS_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef PDF4QTLIBWIDGETS_DEPRECATED
#  define PDF4QTLIBWIDGETS_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef PDF4QTLIBWIDGETS_DEPRECATED_EXPORT
#  define PDF4QTLIBWIDGETS_DEPRECATED_EXPORT PDF4QTLIBWIDGETSSHARED_EXPORT PDF4QTLIBWIDGETS_DEPRECATED
#endif

#ifndef PDF4QTLIBWIDGETS_DEPRECATED_NO_EXPORT
#  define PDF4QTLIBWIDGETS_DEPRECATED_NO_EXPORT PDF4QTLIBWIDGETS_NO_EXPORT PDF4QTLIBWIDGETS_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef PDF4QTLIBWIDGETS_NO_DEPRECATED
#    define PDF4QTLIBWIDGETS_NO_DEPRECATED
#  endif
#endif

#endif /* PDF4QTLIBWIDGETSSHARED_EXPORT_H */
